﻿using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Algorithms.PushGP.Expressions;

namespace HeuristicLab.Algorithms.PushGP
{
    public static class ExpressionTable
    {
        private static Dictionary<string, Expression> statelessExpressionTable;
        private static Dictionary<string, Func<Expression>> statefullExpressionFactory;

        public static int Count;

        static ExpressionTable()
        {
            statelessExpressionTable = GetStatelessExperssionTable(e => e.StringRepresentation);
            statefullExpressionFactory = GetStatefullExpressionFactory(e => e.StringRepresentation);

            Count = statelessExpressionTable.Count + statefullExpressionFactory.Count;
        }

        private static Dictionary<string, Expression> GetStatelessExperssionTable(Func<Expression, string> keyFunc)
        {
            var dictionary = new Dictionary<string, Expression>();
            var expressionTypes = GetExpressionTypes(typeof(StatelessExpression));

            foreach (var type in expressionTypes)
            {
                var expression = Activator.CreateInstance(type) as Expression;
                var key = keyFunc(expression);

                try
                {
                    dictionary.Add(key, expression);
                } catch(Exception e)
                {
                    Console.WriteLine(e);
                }
            }

            return dictionary;
        }

        private static Dictionary<string, Func<Expression>> GetStatefullExpressionFactory(Func<Expression, string> keyFunc)
        {
            var dictionary = new Dictionary<string, Func<Expression>>();
            var expressionTypes = GetExpressionTypes(typeof(StatefullExpression));

            foreach (var type in expressionTypes)
            {
                // Make a NewExpression that calls the ctor
                var newExp = System.Linq.Expressions.Expression.New(type);

                // Create a lambda with the New expression as body
                var creator = System.Linq.Expressions.Expression.Lambda<Func<Expression>>(newExp).Compile();

                var expression = creator();
                var key = keyFunc(expression);

                try
                {
                    dictionary.Add(key, creator);
                }
                catch (Exception e)
                {
                    Console.WriteLine(e);
                }
            }

            return dictionary;
        }

        private static IEnumerable<Type> GetExpressionTypes(Type baseType)
        {
            return from domainAssembly in AppDomain.CurrentDomain.GetAssemblies()
                   from assemblyType in domainAssembly.GetTypes()
                   where !assemblyType.IsAbstract &&
                         assemblyType.IsSubclassOf(baseType) &&
                         assemblyType.GetConstructor(Type.EmptyTypes) != null
                   select assemblyType;
        }

        public static int StatelessCount { get { return statelessExpressionTable.Count; } }

        public static int StatefullCount { get { return statefullExpressionFactory.Count; } }

        public static Expression GetExpression(int index)
        {
            return index < StatelessCount
                ? statelessExpressionTable.ElementAt(index).Value
                : statefullExpressionFactory.ElementAt(index - StatelessCount).Value();
        }

        public static Expression GetExpression(string symbol)
        {
            Expression expression;
            if (TryGetStatelessExpression(symbol, out expression) ||
                TryGetStatefullExpression(symbol, out expression))
            {
                return expression;
            }

            return null;
        }

        public static Expression GetStatelessExpression(string symbol)
        {
            Expression expression;
            if (statelessExpressionTable.TryGetValue(symbol, out expression)) return expression;
            else throw new NotSupportedException("Expression not supported: " + symbol);
        }

        public static bool TryGetStatelessExpression(string name, out Expression expression)
        {
            return statelessExpressionTable.TryGetValue(name, out expression);
        }

        public static Expression GetStatefullExpression(string symbol)
        {
            Func<Expression> creator;
            if (statefullExpressionFactory.TryGetValue(symbol, out creator)) return creator();
            else throw new NotSupportedException("Expression not supported: " + symbol);
        }

        public static bool TryGetStatefullExpression(string name, out Expression expression)
        {
            Func<Expression> creator;
            if (statefullExpressionFactory.TryGetValue(name, out creator))
            {
                expression = creator();
                return true;
            }
            else
            {
                expression = default(Expression);
                return false;
            }

        }
    }
}
